/*
** SpeciesTree.c
*/

# include <stdlib.h>
# include <stdio.h>
# include <string.h>
# include "SpeciesTree.h"

SpeciesTree /*@only@*/ SpeciesTree_new (/*@observer@*/ Species root) {
    SpeciesTree tree = malloc (sizeof (*tree));

    if (tree==NULL) {
      fprintf (stderr, "Out of memory in SpeciesTree_new.\n");
      exit(EXIT_FAILURE);
    }

    tree->root = root;
    tree->left = NULL;
    tree->right = NULL;
    return tree;
}

bool SpeciesTree_isNull (SpeciesTree tree) {
  return tree == SpeciesTree_NULL;
}
  
void SpeciesTree_free (SpeciesTree tree) {
  if (tree != NULL) {
    if (tree->left != NULL) {
      SpeciesTree_free (tree->left);
    }
    
    if (tree->right != NULL) {
      SpeciesTree_free (tree->right);
    }
    
    free (tree);
  }
}

void SpeciesTree_setLeft (SpeciesTree tree,
			  SpeciesTree left) {
  if (tree == NULL) {
    fprintf (stderr, "ERROR: Cannot setLeft on null tree.\n");
    SpeciesTree_free (left);
    return;
  }

  if (tree->left != NULL) {
    SpeciesTree_free (tree->left);
  }

  tree->left = left;
}

void SpeciesTree_setRight (SpeciesTree tree,
			   SpeciesTree right) {
  if (tree == NULL) {
    fprintf (stderr, "ERROR: Cannot setRight on null tree.\n");
    SpeciesTree_free (right);
    return;
  }

  if (tree->right != NULL) {
    SpeciesTree_free(tree->right);
  }

  tree->right = right;
}

SpeciesTree SpeciesTree_getLeft (SpeciesTree tree) {
  if (tree == NULL) {
    fprintf (stderr, "ERROR: Cannot getLeft on null tree.\n");
    return SpeciesTree_NULL;
  }

  return tree->left;
}

SpeciesTree SpeciesTree_getRight (SpeciesTree tree) {
  if (tree == NULL) {
    fprintf (stderr, "ERROR: Cannot getLeft on null tree.\n");
    return SpeciesTree_NULL;
  }

  return tree->right;
}

/*@observer@*/ Species SpeciesTree_getRoot(SpeciesTree tree) {
  if (tree == NULL) {
    fprintf (stderr, "ERROR: Cannot getRoot on null tree.\n");
    exit (EXIT_FAILURE);
  }

  return tree->root;
}

bool SpeciesTree_isEmpty (SpeciesTree tree) {
    return (tree == NULL);
}

static char *SpeciesTree_toStringHelper (SpeciesTree tree, int indent) {
  /* REQUIRES: indent >= 0
   * EFFECTS:  Prints a representation of the tree, indented by
   *	     <indent> spaces.
   */

  int i;
  size_t length;
  char *left;
  char *right;
  char *res;
  char *species;
  char *spaces;

  if (tree == NULL) {
    res = (char *) malloc (sizeof (*res) * (strlen ("<null tree>") + 1));

    if (res == NULL) {
      fprintf (stderr, "Out of memory in SpeciesTree_printHelper.\n");
      exit (EXIT_FAILURE);
    }

    strcpy (res, "<null tree>");
    return res;
  }

  species = Species_toString (tree->root);
  spaces = malloc (sizeof (*spaces) * (indent + 1));

  if (spaces == NULL) {
    fprintf (stderr, "Out of memory in SpeciesTree_printHelper.\n");
    exit (EXIT_FAILURE);
  }

  for (i = 0; i < indent; i++) {
    spaces[i] = ' ';
  }

  spaces[i] = '\0';

  length = strlen (species) + indent + 1;

  if (tree->left != NULL) {
    left = SpeciesTree_toStringHelper (tree->left, indent + 3);
    length += strlen (left) + 1;
  } else {
    left = NULL;
  }

  if (tree->right != NULL) {
    right = SpeciesTree_toStringHelper (tree->right, indent + 3);
    length += strlen (right) + 1;
  } else {
    right = NULL;
  }

  res = malloc (sizeof (*res) * length);

  if (res == NULL) {
    fprintf (stderr, "Out of memory in SpeciesTree_printHelper.  ALlocation length: %d.\n", (int) length);
    exit (EXIT_FAILURE);
  }

  if (left != NULL) {
    if (right != NULL) {
	   // Opps!  Microsoft VC++ doesn't support ISO99 snprintf procedure.
	   // Changed to use _snprintf
      (void) _snprintf (res, length, "%s%s\n%s\n%s", spaces, species, left, right);
      free (right);
    } else {
      (void) _snprintf (res, length, "%s%s\n%s", spaces, species, left);
    } 
    free (left);
  } else {
    if (right != NULL) {
      (void) _snprintf (res, length, "%s%s\n%s", spaces, species, right);
      free (right);
    } else {
      (void) _snprintf (res, length, "%s%s", spaces, species);
    }
  }

  free (species);
  free (spaces);

  return res;
}

char *SpeciesTree_toString (SpeciesTree tree)
{
  return SpeciesTree_toStringHelper (tree, 0);
}
